/**Copyright  2012 held by
 * Robert Williams
 * Robert Poland
 * Marc Hayes
 * Coen Smits
 * Justin Becker
 * All rights reserved.
 * 
 * 
 * This file is part of Bloodmoon.

    Bloodmoon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bloodmoon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bloodmoon.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import javax.imageio.ImageIO;


public class Animator {	
	
	private HashMap<Action, HashMap<Direction, BufferedImage[]>> m_animationMap;
	private HashMap<BodyPart, HashMap<Action, HashMap<Direction, BufferedImage[]>>> m_equipmentAnimationMap;
	private HashMap<BodyPart, Long> m_currentAnimationTimeByPart;
	private Action m_currentAction;
	private Direction m_currentDirection;
	private double m_animationFrequency;
	private long m_currentAnimationTime;
	private long m_currentAnimationMaxDuration;
	private BufferedImage m_currentSpriteSheet;
	private boolean m_loopAnimation;
	
	private String m_debugInfo;

	// ===========================================
	//             Constructors
	// ===========================================
	public Animator(boolean useEquipment) 
	{
		if (useEquipment)
		{
			m_equipmentAnimationMap = new HashMap<BodyPart, HashMap<Action, HashMap<Direction, BufferedImage[]>>>();
			m_currentAnimationTimeByPart = new HashMap<BodyPart, Long>();
			
			for(BodyPart part : BodyPart.values())
			{
				m_currentAnimationTimeByPart.put(part, 0L);
			}
		}

		m_animationMap = new HashMap<Action, HashMap<Direction, BufferedImage[]>>();
		
		m_currentAction = Action.Idle;
		m_animationFrequency = 0.0025;
		m_loopAnimation = true;
	}

	// ===========================================
	//             Public Methods
	// ===========================================

	/**
	 * Add an animation sequence from a sprite sheet
	 * @param act - The action this animation represents
	 * @param dir - The direction this animation faces
	 * @param startX - The x-coordinate of pixel in the top-left corner of the first frame of animation
	 * @param startY - The y-coordinate of pixel in the top-left corner of the first frame of animation
	 * @param frameHeight - The height of each frame, in pixels
	 * @param frameWidth - The width of each frame, in pixels
	 * @param numberOfFrames - The number of frames in the animation sequence
	 */
	public boolean AddAnimation(Action act, Direction dir, int startX, int startY, int frameHeight, int frameWidth, int numberOfFrames)
	{
		if (m_currentSpriteSheet == null)
		{
			return false;
		}
		
		BufferedImage[] animation = new BufferedImage[numberOfFrames];
		
		int row = startY / frameHeight;
		int colOffset = startX / frameWidth;
		
		for (int col = 0; col < numberOfFrames; col++)
		{
			try
			{
				animation[col] = m_currentSpriteSheet.getSubimage(frameWidth * (col + colOffset), frameHeight * row, frameWidth, frameHeight);
			}
			catch(java.awt.image.RasterFormatException ex)
			{
				ex.printStackTrace();
			}
		}
		
		HashMap<Direction, BufferedImage[]> actionMap = m_animationMap.get(act);
		
		if (actionMap == null)
		{
			// Create new map of directions to animations
			actionMap = new HashMap<Direction, BufferedImage[]>();
			
			m_animationMap.put(act, actionMap);
		}
		
		actionMap.put(dir, animation);
		
		return true;
	}
	
	/**
	 * Load an entire sprite sheet's worth of animations, provided each frame is the same size
	 * @param act - The action represented by this sheet
	 * @param directions - Each direction is mapped to a row of frames
	 * @param frameHeight - The height of each frame, in pixels
	 * @param frameWidth- The width of each frame, in pixels
	 * @param numberOfFramesPerRow - The number of columns on the sheet
	 */
	public boolean AddAnimations(Action act, Iterable<Direction> directions, int frameHeight, int frameWidth, int numberOfFramesPerRow)
	{
		boolean success = true;
		int row = 0, col = 0;
		
		for( Direction dir : directions)
		{
			success &= AddAnimation(act, dir, col, row * frameHeight, frameHeight, frameWidth, numberOfFramesPerRow);
			
			row++;
		}
		
		return success;
	}

	/**
	 * Add an animation sequence from a sprite sheet
	 * @param part - The part of the body this is
	 * @param act - The action this animation represents
	 * @param dir - The direction this animation faces
	 * @param startX - The x-coordinate of pixel in the top-left corner of the first frame of animation
	 * @param startY - The y-coordinate of pixel in the top-left corner of the first frame of animation
	 * @param frameHeight - The height of each frame, in pixels
	 * @param frameWidth - The width of each frame, in pixels
	 * @param numberOfFrames - The number of frames in the animation sequence
	 */
	public boolean AddAnimation(BodyPart part, Action act, Direction dir, int startX, int startY, int frameHeight, int frameWidth, int numberOfFrames)
	{
		if (m_currentSpriteSheet == null)
		{
			return false;
		}
		
		BufferedImage[] animation = new BufferedImage[numberOfFrames];
		
		int row = startY / frameHeight;
		int col = startX / frameWidth;
		
		for (; col < numberOfFrames; col++)
		{
			try
			{
				animation[col] = m_currentSpriteSheet.getSubimage(frameWidth * col, frameHeight * row, frameWidth, frameHeight);
			}
			catch(java.awt.image.RasterFormatException ex)
			{
				ex.printStackTrace();
			}
		}
		
		HashMap<Action, HashMap<Direction, BufferedImage[]>> animationMap = m_equipmentAnimationMap.get(part);
		
		if (animationMap == null)
		{
			animationMap = new HashMap<Action, HashMap<Direction, BufferedImage[]>>();
			
			m_equipmentAnimationMap.put(part, animationMap);
		}
		
		HashMap<Direction, BufferedImage[]> actionMap = animationMap.get(act);
		
		if (actionMap == null)
		{
			// Create new map of directions to animations
			actionMap = new HashMap<Direction, BufferedImage[]>();
			
			animationMap.put(act, actionMap);
		}
		
		actionMap.put(dir, animation);
		
		return true;
	}
	
	public boolean AddAnimations(BodyPart part, Action act, Iterable<Direction> directions, int frameHeight, int frameWidth, int numberOfFramesPerRow)
	{
		boolean success = true;
		int row = 0, col = 0;
		
		for( Direction dir : directions)
		{
			success &= AddAnimation(part, act, dir, col, row * frameHeight, frameHeight, frameWidth, numberOfFramesPerRow);
			
			row++;
		}
		
		return success;
	}

	/**
	 * Load a sprite sheet into memory
	 * @param filePath - The relative file path to the sprite sheet
	 * @return True if the image was loaded successfully
	 */
	public boolean LoadSpriteSheet (String filePath)
	{
		File imageFile = new File(filePath);
		try
		{
			m_currentSpriteSheet = ImageIO.read(imageFile);
		}
		catch (IOException e)
		{
			e.printStackTrace();
			
			return false;
		}
		
		return true;
	}
	
	/**
	 * Determine the current frame of animation, using a linear function, based on the current state and time.  
	 * If an animation with a set duration ends, the current action is set to idle.  If an idle animation 
	 * has not been set, this method will return null.
	 * @param elapsedGameTime - The amount of time that has elapsed since the last update, in millisecond 
	 * @return A buffered image that is the frame of animation to be drawn
	 */
	public BufferedImage CalculateNextFrameLinear(long elapsedGameTime)
	{
		// Update the remaining duration of the current animation
		if (m_currentAnimationTime <= 0 && m_loopAnimation)
		{
			m_currentAnimationTime += m_currentAnimationMaxDuration;
			m_currentAnimationTime -= elapsedGameTime;			
		}
		else if (m_currentAnimationTime > 0)
		{
			m_currentAnimationTime -= elapsedGameTime;
		}
		else
		{
			m_currentAction = Action.Idle;
		}
				
		return CalculateNextFrameLinear(m_animationMap);
	}
	
	private BufferedImage CalculateNextFrameLinear(HashMap<Action, HashMap<Direction, BufferedImage[]>> animationMap)
	{	
		BufferedImage img = null;
		
		BufferedImage[] currentAnimationCycle = GetAnimationCycle(animationMap);
		
		if (currentAnimationCycle == null)
		{
			return null;
		}
		
		int totalFrames = currentAnimationCycle.length;
		int currentFrame = 0;

		float ratio = (float)(m_currentAnimationMaxDuration - m_currentAnimationTime) / (float)(m_currentAnimationMaxDuration);
		currentFrame = (int)(ratio * (totalFrames - 1));
		
		if (currentFrame < totalFrames && currentFrame >= 0)
		{
			img = currentAnimationCycle[currentFrame];
		}
		
		m_debugInfo = "Current frame = " + currentFrame + " / " + totalFrames;
		m_debugInfo += " " + "Current action = " + m_currentAction.toString();
		m_debugInfo += " " + "Current direction = " + m_currentDirection.toString();
		
		return img;
	}
	
	public BufferedImage CalculateNextFrameLinear(BodyPart part)
	{
		HashMap<Action, HashMap<Direction, BufferedImage[]>> animationMap = m_equipmentAnimationMap.get(part);
		
		BufferedImage img = CalculateNextFrameLinear(animationMap);
		
		return img;
	}
	
	/**
	 * Determine the current frame of animation, using a sinusoidal function, based on the current state and time.  
	 * If an animation with a set duration ends, the current action is set to idle.  If an idle animation 
	 * has not been set, this method will return null.
	 * @param elapsedGameTime - The amount of time that has elapsed since the last update, in millisecond 
	 * @return A buffered image that is the frame of animation to be drawn
	 */
	public BufferedImage CalculateNextFrameSinusoidal()
	{		
		BufferedImage img = null;
		
		BufferedImage[] currentAnimationCycle = GetAnimationCycle(m_animationMap);
		
		if (currentAnimationCycle == null)
		{
			return null;
		}
		
		long animationTime = System.currentTimeMillis();
		
		int totalFrames = currentAnimationCycle.length;
		int currentFrame = 0;

		currentFrame = (int) Math.abs(totalFrames * Math.sin(m_animationFrequency * animationTime));
		
		if (currentFrame < totalFrames && currentFrame >= 0)
		{
			img = currentAnimationCycle[currentFrame];
		}
		
		m_debugInfo = "Current frame = " + currentFrame + " / " + totalFrames;
		m_debugInfo += " " + "Current action = " + m_currentAction.toString();
		m_debugInfo += " " + "Current animation duration = " + (m_currentAnimationMaxDuration / 1000) + " seconds";
		
		return img;
	}
	
	/**
	 * Set a new state of animation
	 * @param dir - The direction to face for this animation
	 * @param act - The action of this animation
	 * @param duration - How long this animation will last, in milliseconds
	 */
	public void BeginNewAction(Direction dir, Action act, long duration, boolean constantAnimation)
	{
		m_currentDirection = dir;
		m_currentAction = act;
		m_currentAnimationTime = duration;
		m_currentAnimationMaxDuration = duration;
		m_loopAnimation = constantAnimation;
		
		if (m_currentAnimationTimeByPart != null)
		{
			for(BodyPart part : BodyPart.values())
			{
				m_currentAnimationTimeByPart.put(part, duration);
			}
		}
	}
	
	/**
	 * Set the rate at which the animation frames cycle
	 * @param frequency
	 */
	public void SetAnimationFrequency(double frequency)
	{
		m_animationFrequency = frequency;
	}
	
	/**
	 * Dispose of unneeded resources.  This method should be called after all 
	 * animations have been loaded, to avoid unnecessary memory bloat.
	 */
	public void Finish()
	{
		m_currentSpriteSheet = null;
		
		if (m_animationMap.get(Action.Idle) == null)
		{
		}
	}
	
	/**
	 * Get the action currently being animated
	 * @return
	 */
	public Action GetCurrentAction()
	{
		return m_currentAction;
	}
	
	/**
	 * Get the direction currently being animated
	 * @return
	 */
	public Direction GetCurrentDirection()
	{
		return m_currentDirection;
	}
	
	public String GetDebugInfo()
	{
		return m_debugInfo;
	}
	
	// Get the animation cycle for the body, based on current action and direction
	private BufferedImage[] GetAnimationCycle(HashMap<Action, HashMap<Direction, BufferedImage[]>> animationMap)
	{
		if (animationMap == null)
		{
			return null;
		}
		
		// Get all the directions for the current action
		HashMap<Direction, BufferedImage[]> currentActionMap = animationMap.get(m_currentAction);
				
		if (currentActionMap == null)
		{
			return null;
		}
				
		// Get the animation for this action and direction
		BufferedImage[] currentAnimationCycle = currentActionMap.get(m_currentDirection);
				
		return currentAnimationCycle;
	}
}
