/**Copyright  2012 held by
 * Robert Williams
 * Robert Poland
 * Marc Hayes
 * Coen Smits
 * Justin Becker
 * All rights reserved.
 * 
 * 
 * This file is part of Bloodmoon.

    Bloodmoon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bloodmoon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bloodmoon.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.event.KeyEvent;
import java.util.LinkedList;
import java.util.TimerTask;

public class GameEngine extends TimerTask{
	
	// ===========================================
	//             Members
	// ===========================================
	private LinkedList<GameObject> m_currentGameObjects;
	private KeyboardInput m_keyboard;
	private PlayerCharacter [] m_players;
	private LinkedList<Bar> m_bars = null;
	private Bar m_healthBar = null;
	private Bar m_manaBar = null;
	private Bar m_experienceBar = null;
	protected Coin m_uiCoin = null;
	
	LinkedList<GameObject> m_toRemove = new LinkedList<GameObject>();
	LinkedList<GameObject> m_toAdd = new LinkedList<GameObject>();
	
	private Maze m_gameMaze;
	private int m_mazeX;
	private int m_mazeY;

	private double m_mapOffset = 0.0;
	private Direction m_mapMoveDirection = Direction.EAST;
	private double m_mapMoveDelay = 1.0; // Time in seconds to move from one screen to the other
	
	private long m_lastUpdateTime;
	private GameCanvas m_canvas;
	
	private Menu m_menu = null;
	
	private ScrollOverlay m_currentOverlay = null;
	
	private GameMode m_gameMode = GameMode.MENU;

	// Getters
	
	// Setters
	public void setGameMode()
	{
		m_gameMode = GameMode.GAME;
	}
	
	public void setMenuMode()
	{
		m_gameMode = GameMode.MENU;
	}
	
	public void setOverlayMode()
	{
		m_gameMode = GameMode.OVERLAY;
	}
	
	public void setMapMoveMode(Direction direction)
	{
		m_gameMode = GameMode.MAPMOVE;

		// change map tiles
		if(m_mapMoveDirection == Direction.WEST)
		{
			m_mazeX -= 1;
		}
		else if(m_mapMoveDirection == Direction.EAST)
		{
			m_mazeX += 1;
		}
		else if(m_mapMoveDirection == Direction.NORTH)
		{
			m_mazeY -= 1;
		}
		else if(m_mapMoveDirection == Direction.SOUTH)
		{
			m_mazeY += 1;
		}
		
		// spawn monsters for the new map
		int monsterLevel = m_mazeX + m_mazeY + 1;
		SpawnMonsters(monsterLevel);
		
	}
	
	// ===========================================
	//             Constructors
	// ===========================================
	public GameEngine(int numberOfPlayers, KeyboardInput keyboard)
	{
		m_menu = new Menu(this);
		
		m_keyboard = keyboard;
		
		// initialize lists
		m_currentGameObjects = new LinkedList<GameObject>();
		m_bars = new LinkedList<Bar>();
		
		// character
		m_players = new PlayerCharacter[numberOfPlayers];
		for (int i = 0; i < m_players.length; i++)
		{
			m_players[i] = new PlayerCharacter();
		}
		
		// Bars
		int barWidth = 233;
		int offset = 25;
		m_healthBar = new Bar(offset, 10, (int)m_players[0].GetHealth(), m_players[0].GetMaxHealth(), Color.RED);
		m_bars.add(m_healthBar);
		
		m_manaBar = new Bar(offset + barWidth, 10, m_players[0].GetMana(), m_players[0].GetMaxMana(), Color.BLUE);
		m_bars.add(m_manaBar);

		m_experienceBar = new Bar(offset + 2 * barWidth, 10, m_players[0].m_exp, m_players[0].m_expToLevel, Color.GREEN.darker());
		m_bars.add(m_experienceBar);
		
		// ui lists
		m_uiCoin = new Coin(Coin.CoinType.GOLD, offset + 3 * barWidth, 10);
		m_uiCoin.m_permanent = true;
		m_currentGameObjects.add(m_uiCoin);
		
		// load in maps
		m_gameMaze = new Maze(".\\Resources\\Maps\\Maze1.maze");
		m_mazeX = m_mazeY = 0;
		
		// zone transitions
		AddDefaultZoneBlocks();
				
		m_lastUpdateTime = System.currentTimeMillis();
		
		ShowHelp();
	}

	// ==============================================
	//              Public Methods
	// ==============================================
	
	/**
	 * During a repaint, call on all known objects to draw themselves
	 * @param g
	 */
	public void PaintGameStateToScreen(Graphics2D graphics)
	{
		//g.drawImage(m_map, 0, 0, m_map.getWidth(), m_map.getHeight(), Color.white, null);
		
		if( m_gameMode == GameMode.MAPMOVE )
		{
			// Render both maps, but with an offset
			RenderMapMovement(graphics);
			return;
		}
		
		//m_sandbox.DrawMap(graphics);
		m_gameMaze.m_zone[m_mazeX][m_mazeY].DrawMap(graphics);
		
//		graphics.drawString(String.valueOf(m_mazeX), 100, 100);
//		graphics.drawString(String.valueOf(m_mazeY), 100, 120);
		
		for(GameObject object : m_currentGameObjects)
		{
			// de-spawn blinking
			if (object instanceof Coin)
			{
				if(((Coin) object).m_lifeSpan < 3000 && !((Coin) object).m_permanent)
				{
					if(Math.sin(100 * System.currentTimeMillis()) > 0)
					{
						object.Paint(graphics);
					}
				}
				else
				{
					object.Paint(graphics);
				}
			}
			else if(object instanceof Item)
			{
				if(((Item) object).m_lifeSpan < 3000)
				{
					if(Math.sin(100 * System.currentTimeMillis()) > 0)
					{
						object.Paint(graphics);
					}
				}
				else
				{
					object.Paint(graphics);
				}
			}
			// normal drawing
			else
			{
				object.Paint(graphics);
			}
		}
		
		for(PlayerCharacter playerCharacter : m_players)
		{
			playerCharacter.Paint(graphics);
		}
		
		for(Bar bar : m_bars)
		{
			bar.Paint(graphics);
		}
		
		if( m_gameMode == GameMode.MENU )
		{
			m_menu.Paint(graphics);
		}
		
		if( m_gameMode == GameMode.OVERLAY )
		{
			// Render overlay
			m_currentOverlay.Paint(graphics);
		}
		
		for(Wall w : m_gameMaze.m_zone[m_mazeX][m_mazeY].m_walls)
		{
			w.Paint(graphics);
		}
	}
	
	public void SpawnMonsters(int monsterLevel)
	{
		// clear the list of all current monsters
		for(GameObject obj : m_currentGameObjects)
		{
			if(obj instanceof Monster)
			{
				m_toRemove.add(obj);
			}
		}
		
		// add in new monsters
		if(monsterLevel > 8)
		{
			//Boss room
			m_toAdd.add(new Pumpking(m_players[0].GetLevel() + 2));
			// TODO: Boss Room - make fancy
		}
		for(int i = 0; i < (monsterLevel / 2); i++)
		{
			m_toAdd.add(new Bat(monsterLevel));
		}
	}

	
	/**
	 * Set a canvas for the engine to invoke a repaint on
	 * @param canvas
	 */
	public void SetCanvas(GameCanvas canvas)
	{
		m_canvas = canvas;
	}

	public int GetMapWidth()
	{
		return m_gameMaze.m_zone[m_mazeX][m_mazeY].GetMapWidth();
	}
	
	public int GetMapHeight()
	{
		return m_gameMaze.m_zone[m_mazeX][m_mazeY].GetMapHeight();
	}
	
	// ==============================================
	//                   Overrides
	// ==============================================
	@Override
	public void run()
	{		
		// Main game loop for calculating the game state.
		long currentTime = System.currentTimeMillis();
		long elapsedTime = currentTime - m_lastUpdateTime;
		m_lastUpdateTime = currentTime;
		
		m_keyboard.poll();

		if( m_gameMode == GameMode.MENU )
		{
			m_menu.Update(elapsedTime, m_keyboard);		
		}
		else if( m_gameMode == GameMode.OVERLAY )
		{
			m_currentOverlay.Update(elapsedTime, m_keyboard);
		}
		else if( m_gameMode == GameMode.MAPMOVE )
		{
			CalculateMapMovement(elapsedTime);
		}
		else if (m_gameMode == GameMode.GAME_OVER)
		{
			DisplayGameOver();
		}
		else if (m_gameMode == GameMode.GAME_VICTORY)
		{
			DisplayVictory();
		}
		else
		{			
			if( m_keyboard.keyDownOnce(KeyEvent.VK_ESCAPE) )
			{
				setMenuMode();
				return;
			}
			
			if( m_keyboard.keyDownOnce(KeyEvent.VK_I) ||
				m_keyboard.keyDownOnce(KeyEvent.VK_TAB) )
			{
				// Show inventory
				ShowInventory();
				return;
			}
			else if (m_keyboard.keyDownOnce(KeyEvent.VK_U) ||
					 m_keyboard.keyDownOnce(KeyEvent.VK_R) ||
					 m_players[0].m_skillPoints > 0)
			{
				ShowSkills();
				return;
			}
			else if (m_keyboard.keyDownOnce(KeyEvent.VK_H))
			{
				ShowHelp();
			}
			
			// loop over object list and call update
			for(PlayerCharacter player :  m_players)
			{
				player.Update(elapsedTime);
				player.Update(elapsedTime, m_keyboard, m_currentGameObjects);
			}
			
			m_toRemove = new LinkedList<GameObject>();
			m_toAdd = new LinkedList<GameObject>();
			
			// loop over objects for updates
			for(GameObject obj : m_currentGameObjects)
			{
				// update position/action
				obj.Update(elapsedTime);
				
				// check for collisions
				for(PlayerCharacter player : m_players)
				{
					player.CheckCollisions(obj);
					
					obj.CheckCollisions(player);
					
					// bound checking for walls
					for(Wall w : m_gameMaze.m_zone[m_mazeX][m_mazeY].m_walls)
					{
						player.CheckCollisions(w);
					}
				}
				
				if (obj instanceof Monster)
				{
					Monster monster = (Monster)obj;

					// bound checking for walls
					for(Wall w : m_gameMaze.m_zone[m_mazeX][m_mazeY].m_walls)
					{
						monster.CheckCollisions(w);
					}
					
					// check if monsters died
					if (monster.MonsterDead())
					{
						// remove monster
						m_toRemove.add(monster);
						
						if (monster instanceof Pumpking)
						{
							m_gameMode = GameMode.GAME_VICTORY;
						}
					
						// get loot for dead monster
						GameObject loot = monster.GetLoot();
						if (loot != null)
						{
							m_toAdd.add(loot);
						}
						
						// deal with exp for dead monsters
						m_players[0].getExp(monster);
					}
					
					if( monster.IsClose(m_players[0]) )
					{
						// Get into aggressive mode to hunt for player
						monster.setAggressoMode(true);
					}
					else
					{
						monster.setAggressoMode(false);
					}
				}
				
				if (obj.IsExpired())
				{
					m_toRemove.add(obj);
				}

				if (m_players[0].GetHealth() <= 0)
				{
					m_gameMode = GameMode.GAME_OVER;
				}
			}
			
			// update all bars
			m_healthBar.setCurrentValue((int)m_players[0].GetHealth(), m_players[0].GetMaxHealth());
			m_manaBar.setCurrentValue(m_players[0].GetMana(), m_players[0].GetMaxMana());
			m_experienceBar.setCurrentValue(m_players[0].m_exp, m_players[0].m_expToLevel);

	
			// remove dead monsters
			for(GameObject deadMonster : m_toRemove)
			{
				m_currentGameObjects.remove(deadMonster);
			}
	
			// respawns
			for(GameObject m : m_toAdd)
			{
				m_currentGameObjects.add(m);
			}
		}
		
		// redraw screen after all updates
		if (m_canvas != null)
		{
			m_canvas.repaint();
		}
	}
	
	public void ChangeZone(Direction direction)
	{
		m_mapMoveDirection = direction;
		
		// Set the position of the character
		for(PlayerCharacter playerCharacter : m_players)
		{
			Vector2D playerPosition = playerCharacter.getPosition();
			
			if( direction == Direction.NORTH )
			{
				int i = (int)(playerPosition.GetX() / Map.TILESIZE);
				playerCharacter.setPosition(i, 17);
			}
			else if( direction == Direction.EAST )
			{
				int j = (int)(playerPosition.GetY() / Map.TILESIZE);
				playerCharacter.setPosition(3, j);
			}
			else if( direction == Direction.SOUTH )
			{
				int i = (int)(playerPosition.GetX() / Map.TILESIZE);
				playerCharacter.setPosition(i, 3);
			}
			else if( direction == Direction.WEST )
			{
				int j = (int)(playerPosition.GetY() / Map.TILESIZE);
				playerCharacter.setPosition(26,j);
			}
			
			playerCharacter.m_boundingBox.setBounds((int)(playerCharacter.m_position.GetX()) + playerCharacter.m_boundingBoxOffSetX,
													(int)(playerCharacter.m_position.GetY()) + playerCharacter.m_boundingBoxOffSetY,
													playerCharacter.m_boundingBox.width,
													playerCharacter.m_boundingBox.height);
		}
		
		// Set the game mode to map move
		setMapMoveMode(direction);
 	}
	
	private void CalculateMapMovement(long elapsedTime)
	{
		m_mapOffset += elapsedTime / (1000.0 * m_mapMoveDelay);
		
		if( m_mapOffset >= 1.0)
		{
			// Stop moving
			setGameMode();
			m_mapOffset = 0.0;
		}
	}
	
	private void RenderMapMovement(Graphics2D graphics)
	{
		int maxWidth = graphics.getClip().getBounds().width;
		int maxHeight = graphics.getClip().getBounds().height;

		int offsetX = 0;
		int prevMapOffsetX = 0;
		int offsetY = 0;
		int prevMapOffsetY = 0;
	
		int oldX = m_mazeX;
		int oldY = m_mazeY;
		
		if(m_mapMoveDirection == Direction.WEST)
		{
			oldX += 1;
			offsetX = (int)(m_mapOffset * maxWidth);
			prevMapOffsetX = offsetX - maxWidth;
			m_gameMaze.m_zone[m_mazeX][m_mazeY].DrawMap(graphics, prevMapOffsetX, prevMapOffsetY);
			m_gameMaze.m_zone[oldX][oldY].DrawMap(graphics, offsetX, offsetY);
		}
		else if(m_mapMoveDirection == Direction.EAST)
		{
			oldX -= 1;
			offsetX = (int)((1.0 - m_mapOffset) * maxWidth);
			prevMapOffsetX = offsetX - maxWidth;
			m_gameMaze.m_zone[oldX][oldY].DrawMap(graphics, prevMapOffsetX, prevMapOffsetY);
			m_gameMaze.m_zone[m_mazeX][m_mazeY].DrawMap(graphics, offsetX, offsetY);		
		}
		else if(m_mapMoveDirection == Direction.NORTH)
		{
			oldY += 1;
			offsetY = (int)(m_mapOffset * maxHeight);
			prevMapOffsetY = offsetY - maxHeight;
			m_gameMaze.m_zone[m_mazeX][m_mazeY].DrawMap(graphics, prevMapOffsetX, prevMapOffsetY);
			m_gameMaze.m_zone[oldX][oldY].DrawMap(graphics, offsetX, offsetY);
		}
		else if(m_mapMoveDirection == Direction.SOUTH)
		{
			oldY -= 1;
			offsetY = (int)((1.0 - m_mapOffset) * maxHeight);
			prevMapOffsetY = offsetY - maxHeight;
			m_gameMaze.m_zone[oldX][oldY].DrawMap(graphics, prevMapOffsetX, prevMapOffsetY);
			m_gameMaze.m_zone[m_mazeX][m_mazeY].DrawMap(graphics, offsetX, offsetY);		
		}
	}
	
	private void ShowInventory()
	{
		PlayerCharacter player = m_players[0];
		
		m_currentOverlay = new InventoryOverlay(this, player);
	
		setOverlayMode();
	}
	
	private void ShowSkills()
	{
		PlayerCharacter player = m_players[0];
		
		m_currentOverlay = new PlayerSkillOverlay(this, player);
	
		setOverlayMode();
	}
	
	private void DisplayGameOver()
	{
		m_currentOverlay = new GameOverScroll(this, GameMode.GAME_OVER);
	
		setOverlayMode();
	}
	
	private void DisplayVictory()
	{
		m_currentOverlay = new GameOverScroll(this, GameMode.GAME_VICTORY);
	
		setOverlayMode();
	}
	public void ShowHelp()
	{
		m_currentOverlay = new HelpOverlay(this);
	
		setOverlayMode();
	}
	
	private void AddDefaultZoneBlocks()
	{
		MapZoneBlock mapZoneBlockNorth = new MapZoneBlock(this, Direction.NORTH);
		mapZoneBlockNorth.setBoundingBox(13, 0, 4, 1);
		
		MapZoneBlock mapZoneBlockEast = new MapZoneBlock(this, Direction.EAST);
		mapZoneBlockEast.setBoundingBox(29,9,1, 4);
		
		MapZoneBlock mapZoneBlockSouth = new MapZoneBlock(this, Direction.SOUTH);
		mapZoneBlockSouth.setBoundingBox(13,21,4, 1);
		
		MapZoneBlock mapZoneBlockWest = new MapZoneBlock(this, Direction.WEST);
		mapZoneBlockWest.setBoundingBox(0,9,1, 4);

		m_currentGameObjects.add(mapZoneBlockNorth);
		m_currentGameObjects.add(mapZoneBlockEast);
		m_currentGameObjects.add(mapZoneBlockSouth);
		m_currentGameObjects.add(mapZoneBlockWest);
	}
}
