/**Copyright  2012 held by
 * Robert Williams
 * Robert Poland
 * Marc Hayes
 * Coen Smits
 * Justin Becker
 * All rights reserved.
 * 
 * 
 * This file is part of Bloodmoon.

    Bloodmoon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bloodmoon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bloodmoon.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;

public abstract class GameObject {

	// ==============================================
	//                   Members
	// ==============================================
	
	/**
	 * Image representing what this object looks like
	 */
	protected BufferedImage m_currentImage;	
	protected Direction m_facingDirection;
	
	protected Animator m_animator = null;
	
	protected Vector2D m_position = new Vector2D(0,0);
	protected Vector2D m_velocity = new Vector2D(0.0f, 0.0f);
	protected Vector2D m_acceleration = new Vector2D(0.0f, 0.0f); 
	
	protected Rectangle m_boundingBox;
	protected int m_boundingBoxOffSetX = 0;
	protected int m_boundingBoxOffSetY = 0;
	
	protected float m_deceleration = 0.0f;
	protected boolean m_usesAccel = false;
	protected boolean m_usesDecel = false;
	protected boolean m_usesBounding = true;
	
	private boolean m_expired = false;
		
	private String m_debugInfo = "";
	private boolean m_showDebug = false;
	private Color m_debugColor = Color.black;
	
	// used for monsters that 'float' with an animation
	protected boolean m_animationIsConstant; 
	
	// Getters
	public Vector2D getPosition()
	{
		return m_position;
	}
	
	// Setters
	public void setBoundingBox(Rectangle value)
	{
		m_boundingBox = value;
	}
	
	public void setBoundingBox(int startI, int startJ, int nI, int nJ)
	{
		Rectangle boundingBoxRectangle = new Rectangle();
		boundingBoxRectangle.x = startI * Map.TILESIZE;
		boundingBoxRectangle.y = startJ  * Map.TILESIZE;
		boundingBoxRectangle.width = nI * Map.TILESIZE;
		boundingBoxRectangle.height = nJ * Map.TILESIZE;
		
		setBoundingBox(boundingBoxRectangle);
	}
	
	public void setPosition(Vector2D value)
	{
		m_position.SetX(value.GetX());
		m_position.SetY(value.GetY());
	}
	
	public void setPosition(int i, int j)
	{
		Vector2D position = new Vector2D(Map.TILESIZE * i, Map.TILESIZE * j);
		setPosition(position);
	}
	
	// ==============================================
	//                 Constructors
	// ==============================================

	public GameObject()
	{
		m_animator = new Animator(false);
		
		Initialize();
		
		m_animator.Finish();
	}
	

	// ==============================================
	//                Abstract Methods
	// ==============================================
	/**
	 * Update the GameObject based on the elapsed time since last update
	 * @param elapsedGameTime
	 */
	public abstract void Update(long elapsedGameTime);

	/**
	 * Load all sprite animations
	 */
	protected abstract void InitializeSprites();
	
	/**
	 * Initialize starting state vectors
	 */
	protected abstract void InitializeStartingVectors();
	

	// ==============================================
	//                Public Methods
	// ==============================================
	/**
	 * Paint the GameObject to the screen
	 * @param g
	 */
	public void Paint(Graphics2D g)
	{
		if (m_currentImage == null)
		{
			return;
		}
		
		int x = (int)m_position.GetX();
		int y = (int)m_position.GetY();
		int width = m_currentImage.getWidth();
		int height = m_currentImage.getHeight();
		
		if (m_currentImage != null)
		{
			g.drawImage(m_currentImage, x, y, width, height, null, null);
		}
		
		PaintSpellEffects(g);
	}
	
	/**
	 * Override this method to handle collisions with other objects
	 * @param obj
	 */
	public void CheckCollisions(GameObject obj)
	{		
	}
	
	/**
	 * Get the bounding box for this object.  Use this rectangle to check 
	 * for collisions
	 * @return
	 */
	public Rectangle GetBoundingBox()
	{
		return m_boundingBox;
	}	

	/**
	 * Set this object as expired, which will flag it for removal from the 
	 * game state
	 */
	public void SetExpired()
	{
		m_expired = true;
	}
	
	/**
	 * If this object is expired, it is no longer a valid object and should 
	 * be ignored by other objects.  It will be removed from the game state.
	 * @return
	 */
	public boolean IsExpired()
	{
		return m_expired;
	}
	
	public float GetPositionX()
	{
		return m_position.GetX();
	}
	
	public float GetPositionY()
	{
		return m_position.GetY();
	}
	
	// ==============================================
	//              Protected Methods
	// ==============================================

	protected void InitializeBoundingBox()
	{
		if (m_position == null || m_currentImage == null)
		{
			return;
		}
		
		m_boundingBox = new Rectangle((int)m_position.GetX(), (int)m_position.GetY(), m_currentImage.getWidth(), m_currentImage.getHeight());
		
		m_boundingBoxOffSetX = 0;
		m_boundingBoxOffSetY = 0;
	}
	
	protected void PaintSpellEffects(Graphics2D g)
	{
		if (m_showDebug)
		{
			g.setColor(m_debugColor);
			
			if (m_boundingBox != null)
			{
				g.drawRect(m_boundingBox.x, m_boundingBox.y, m_boundingBox.width, m_boundingBox.height);
			}			
			
			int posX = (int) m_boundingBox.x - (m_debugInfo.length() / 2);
			int posY = (int) m_boundingBox.y + m_boundingBox.height + 10;
			
			g.drawString(m_debugInfo, posX, posY);
		}
	}
		
	protected void Initialize()
	{
		InitializeSprites();
		InitializeStartingVectors();
		InitializeBoundingBox();
	}
	
	protected void UpdatePosition(long elapsedGameTime) //, LinkedList<GameObject> currentGameObjects)
	{
		//Vector2D newPosition = new Vector2D(m_position.GetX(), m_position.GetY());

		// apply accel if needed
		if(m_usesAccel == true)
		{
			m_velocity.SetX(elapsedGameTime * m_acceleration.GetX());
			m_velocity.SetY(elapsedGameTime * m_acceleration.GetY());
		}
		
		// update position (possible)
		m_position.AddX(elapsedGameTime * m_velocity.GetX());
		m_position.AddY(elapsedGameTime * m_velocity.GetY());
		
		// apply decel if needed
		if(m_usesDecel == true)
		{
			m_velocity.SetX(m_velocity.GetX() * m_deceleration);
			m_velocity.SetY(m_velocity.GetY() * m_deceleration);
			
			m_acceleration.SetX(m_velocity.GetX() * m_deceleration);
			m_acceleration.SetY(m_velocity.GetY() * m_deceleration);
		}

		// bound check the new position
//		if(m_usesBounding == true)
//		{
//			for(GameObject g : currentGameObjects)
//			{
//				if(this.CheckCollisions(g));
//			}
//		}
		
		// if failed, stop movement
		

		// update bounding box to match new position
		m_boundingBox.setLocation((int)m_position.GetX() + m_boundingBoxOffSetX, (int)m_position.GetY() + m_boundingBoxOffSetY);
	}
	
	// Debug Methods.  
	// TODO REMOVE BEFORE RELEASE
	protected void SetShowDebug(boolean showDebug)
	{
		m_showDebug = showDebug;
	}
	
	protected void SetDebugInfo(String info)
	{
		m_debugInfo = info;
	}
	
	protected void AppendDebugInfo(String info)
	{
		m_debugInfo += " " + info;
	}
	
	protected void SetDebugColor(Color c)
	{
		m_debugColor = c;
	}
}
