/**Copyright  2012 held by
 * Robert Williams
 * Robert Poland
 * Marc Hayes
 * Coen Smits
 * Justin Becker
 * All rights reserved.
 * 
 * 
 * This file is part of Bloodmoon.

    Bloodmoon is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Bloodmoon is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Bloodmoon.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.LinkedList;

import javax.imageio.ImageIO;


public class Map {

	// ==============================================
	//                 Members
	// ==============================================
	static protected int TILESIZE = 32;
	
	protected BufferedImage[] m_tiles;
	protected int[][] m_map;

	protected int m_mapWidth;
	protected int m_mapHeight;
	
	protected int m_tileHeight = TILESIZE;
	protected int m_tileWidth = TILESIZE;
	
	// is there a door in that direction
	protected boolean m_zoneN;
	protected boolean m_zoneS;
	protected boolean m_zoneE;
	protected boolean m_zoneW;

	// list of bounding walls for collision
	protected LinkedList<Wall> m_walls;


	// ==============================================
	//                 Constructors
	// ==============================================
	public Map() {

	}
	
	public Map(String filePathToImages, int roomStyle)
	{
		m_tiles = new BufferedImage[15];
		
		BufferedImage spriteSheet = null;
		
		// Load the texture file
		try
		{
			BufferedReader textureFile = new BufferedReader(new FileReader(filePathToImages));
			m_map = new int[m_mapWidth][m_mapHeight];
			
			int sheetWidth = 3;
			int sheetHeight = 5;
			// Load in the sprite sheet into the tile set
			File imageFile = new File(filePathToImages);
			spriteSheet = ImageIO.read(imageFile);
			for(int j = 0; j < sheetHeight; j++)
			{
				for(int i = 0; i < sheetWidth; i++)
				{
					m_tiles[(j*sheetWidth) + i] = spriteSheet.getSubimage(i * m_tileWidth, j * m_tileHeight, m_tileWidth, m_tileHeight);
				}
			}
			
			// close the file
			textureFile.close();
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}

		// set doors based on room style
		switch(roomStyle)
		{
		case 1: // open: top
			m_zoneN = true;
			m_zoneS = false;
			m_zoneE = false;
			m_zoneW = false;
			break;
		case 2: // open: right
			m_zoneN = false;
			m_zoneS = false;
			m_zoneE = true;
			m_zoneW = false;
			break;
		case 3: // open: bottom
			m_zoneN = false;
			m_zoneS = true;
			m_zoneE = false;
			m_zoneW = false;
			break;
		case 4: // open: left
			m_zoneN = false;
			m_zoneS = false;
			m_zoneE = false;
			m_zoneW = true;
			break;
		case 5: // open: top, right
			m_zoneN = true;
			m_zoneS = false;
			m_zoneE = true;
			m_zoneW = false;
			break;
		case 6: // open: top, bottom
			m_zoneN = true;
			m_zoneS = true;
			m_zoneE = false;
			m_zoneW = false;
			break;
		case 7: // open: top, left
			m_zoneN = true;
			m_zoneS = false;
			m_zoneE = false;
			m_zoneW = true;
			break;
		case 8: // open: right, bottom
			m_zoneN = false;
			m_zoneS = true;
			m_zoneE = true;
			m_zoneW = false;
			break;
		case 9: // open: left, right
			m_zoneN = false;
			m_zoneS = false;
			m_zoneE = true;
			m_zoneW = true;
			break;
		case 10: // open: left, bottom
			m_zoneN = false;
			m_zoneS = true;
			m_zoneE = true;
			m_zoneW = false;
			break;
		case 11: // open: top, right, bottom
			m_zoneN = true;
			m_zoneS = true;
			m_zoneE = true;
			m_zoneW = false;
			break;
		case 12: // open: top, left, bottom
			m_zoneN = true;
			m_zoneS = true;
			m_zoneE = false;
			m_zoneW = true;
			break;
		case 13: // open: top, left, right
			m_zoneN = true;
			m_zoneS = false;
			m_zoneE = true;
			m_zoneW = true;
			break;
		case 14: // open: left, right, bottom
			m_zoneN = false;
			m_zoneS = true;
			m_zoneE = true;
			m_zoneW = true;
			break;
		case 15: // open: top, left, right, bottom
			m_zoneN = true;
			m_zoneS = true;
			m_zoneE = true;
			m_zoneW = true;
			break;
		}
		
		m_mapWidth = 30;
		m_mapHeight = 22;
		
		m_map = new int[m_mapWidth][m_mapHeight];
		// random grass for start
		for(int i = 0; i < m_mapWidth; i++)
		{
			for(int j = 0; j < m_mapHeight; j++)
			{
				m_map[i][j] = Global.rand.nextInt(2) * 3;
			}
		}
		
		// create the tile map for the outside
		for(int i = 0; i < m_mapWidth; i++)
		{
			m_map[i][0] = 10;
			m_map[i][1] = 13;
			m_map[i][20] = 7;
			m_map[i][21] = 10;
		}
		for(int i = 1; i < m_mapHeight - 1; i++)
		{
			m_map[0][i] = 10;
			m_map[1][i] = 11;
			m_map[28][i] = 9;
			m_map[29][i] = 10;
		}
		
		// inside corner
		m_map[1][1] = 1;
		m_map[1][20] = 4;
		m_map[28][1] = 2;
		m_map[28][20] = 5;
		
		// overwrite the tiles with the door code if open
		if(m_zoneN)
		{
			m_map[12][0] = 11;
			m_map[13][0] = 0;
			m_map[14][0] = 0;
			m_map[15][0] = 0;
			m_map[16][0] = 0;
			m_map[17][0] = 9;
			m_map[12][1] = 14;
			m_map[13][1] = 0;
			m_map[14][1] = 0;
			m_map[15][1] = 0;
			m_map[16][1] = 0;
			m_map[17][1] = 12;
		}
		if(m_zoneS)
		{
			m_map[12][20] = 8;
			m_map[13][20] = 0;
			m_map[14][20] = 0;
			m_map[15][20] = 0;
			m_map[16][20] = 0;
			m_map[17][20] = 6;
			m_map[12][21] = 11;
			m_map[13][21] = 0;
			m_map[14][21] = 0;
			m_map[15][21] = 0;
			m_map[16][21] = 0;
			m_map[17][21] = 9;
		}
		if(m_zoneE)
		{
			m_map[28][8] = 12;
			m_map[28][9] = 0;
			m_map[28][10] = 0;
			m_map[28][11] = 0;
			m_map[28][12] = 0;
			m_map[28][13] = 6;
			m_map[29][8] = 13;
			m_map[29][9] = 0;
			m_map[29][10] = 0;
			m_map[29][11] = 0;
			m_map[29][12] = 0;
			m_map[29][13] = 7;
		}
		if(m_zoneW)
		{
			m_map[0][8] = 13;
			m_map[0][9] = 0;
			m_map[0][10] = 0;
			m_map[0][11] = 0;
			m_map[0][12] = 0;
			m_map[0][13] = 7;
			m_map[1][8] = 14;
			m_map[1][9] = 0;
			m_map[1][10] = 0;
			m_map[1][11] = 0;
			m_map[1][12] = 0;
			m_map[1][13] = 8;
		}
		
		
		// create the list of walls that are collidable
		m_walls = new LinkedList<Wall>();
		// vertical walls
		for(int i = 0; i < 8; i++)
		{
			m_walls.add(new Wall(1, i + 1));
			m_walls.add(new Wall(1, i + 13));
			m_walls.add(new Wall(28, i + 1));
			m_walls.add(new Wall(28, i + 13));
		}
		// horizontal walls
		for(int i = 0; i < 11; i++)
		{
			m_walls.add(new Wall(i + 2, 1));
			m_walls.add(new Wall(i + 17, 1));
			m_walls.add(new Wall(i + 2, 20));
			m_walls.add(new Wall(i + 17, 20));
		}
		// doorways
		for(int i = 0; i < 4; i++)
		{
			if(!m_zoneN)
			{
				m_walls.add(new Wall(i + 13, 1));
			}
			if(!m_zoneS)
			{
				m_walls.add(new Wall(i + 13, 20));
			}
			if(!m_zoneE)
			{
				m_walls.add(new Wall(28, i + 9));
			}
			if(!m_zoneW)
			{
				m_walls.add(new Wall(1, i + 9));
			}
		}
	}
	
	/**
	 * Loads a map from a file provided. File must be in the set style, example is "example.map"
	 * @param FilePath
	 */
	public Map(String filePathToMap) 
	{
		m_tiles = new BufferedImage[3];
		
		String filePathToImages = null;
		BufferedImage spriteSheet = null;
		
		int textureMapTilesWidth;
		int textureMapTilesHeight;
		
		// Load the map file
		try
		{
		BufferedReader mapFile = new BufferedReader(new FileReader(filePathToMap));
			// Load all data from the map file
			// Texture Map Details
			// -Path
			filePathToImages = mapFile.readLine();
					
			// -number of tiles width/height
			textureMapTilesWidth = Integer.parseInt(mapFile.readLine());
			textureMapTilesHeight = Integer.parseInt(mapFile.readLine());

			// -pixel size of tiles width/height
			m_tileWidth = Integer.parseInt(mapFile.readLine());
			m_tileHeight = Integer.parseInt(mapFile.readLine());
			
			// Map Details
			// -size of map width/height (in tiles)
			m_mapWidth = Integer.parseInt(mapFile.readLine());
			m_mapHeight = Integer.parseInt(mapFile.readLine());
			
			m_map = new int[m_mapWidth][m_mapHeight];
			
			// Load in the sprite sheet into the tile set
			File imageFile = new File(filePathToImages);
			spriteSheet = ImageIO.read(imageFile);
			for(int j = 0; j < textureMapTilesHeight; j++)
			{
				for(int i = 0; i < textureMapTilesWidth; i++)
				{
					m_tiles[(j*textureMapTilesWidth) + i] = spriteSheet.getSubimage(i * m_tileWidth, j * m_tileHeight, m_tileWidth, m_tileHeight);
				}
			}
			
			// -matrix for the map of all indexes
			for(int j = 0; j < m_mapHeight; j++)
			{
				String s[] = mapFile.readLine().split(" ");
				for(int i = 0; i < m_mapWidth; i++)
				{
					m_map[i][j] = Integer.parseInt(s[i]);
				}
			}

			// TODO: add in special objects here like trees
			int numberOfSpecials = Integer.parseInt(mapFile.readLine());
			for(int i = 0; i < numberOfSpecials; i++)
			{
				// Read in special objects and parse
			}
			
			// close the file
			mapFile.close();
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
	}

	// ==============================================
	//                 Public Methods
	// ==============================================
	public void DrawMap(Graphics2D graphics)
	{
		DrawMap(graphics, 0, 0);
	}
		
	public void DrawMap(Graphics2D graphics, int offsetX, int offsetY)
	{
		for(int i = 0; i < m_mapWidth; i++)
		{
			for (int j = 0; j < m_mapHeight; j++)
			{
				int tileIndex = m_map[i][j];

				graphics.drawImage(m_tiles[tileIndex], null, offsetX + (i * m_tileWidth), offsetY + (j * m_tileHeight));
			}
		}
	}
	
	public int GetMapHeight()
	{
		return m_mapHeight * m_tileHeight;
	}
	
	public int GetMapWidth()
	{
		return m_mapWidth * m_tileWidth;
	}

	// ==============================================
	//               Private Methods
	// ==============================================
	
}
